* Program..: GetCol.prg
* Author...: Jeremy Suiter
* Date.....: 19/03/2003
*
* Notice...: Copyright (c) 2003, Jeremy Suiter, All Rights Reserved
* Notes....: Visual DbEditor Get column functions (modified from ATSAG samples)

#pragma Library("XppUI2.LIB")

#include "Appevent.ch"
#include "Common.ch"

CLASS XbpGetColumn FROM XbpColumn
   PROTECTED:
   CLASS VAR activeColumn SHARED // The column that is currently being edited
   VAR    xbpGet                 // The associated XbpGet object
   METHOD createXbp              // Create it upon first edit request
   METHOD setPosAndSize          // Assign pos and size for editing

   EXPORTED:
   INLINE METHOD preValidate     // These methods are called by an
   Return .T.                    // XbpGet object during editing.
                                 // We don't do validation in a browser
   INLINE METHOD postValidate    // so these are dummy methods that
   Return .T.                    // must Return a logical value.

   VAR setHome                   // Used by an XbpGet object

                                 ** Overloaded methods
   METHOD init                   // Initialize object
   METHOD create                 // Request system resources
   METHOD destroy                // Release System resources
   METHOD resize                 // Resize must be reflected by XbpGet during edit
   METHOD hiliteRow              // Display the XbpGet during browsing

                                 ** This is required for editing
   VAR    active  READONLY       // Tells If we are in edit mode
   VAR    browser                // XbpBrowse object the column is contained in
   VAR    bufferLength           // Max. number of characters for editing
   VAR    picture                // picture cluase for XbpGet

   METHOD _lbDblClick            // lbDblClick method for :dataArea object
   METHOD read                   // Begin editing
   METHOD keyBoard               // Called by an XbpGet object for UP / DOWN keys
   METHOD killRead               // End editing
ENDCLASS


/*
 * Initialize the object
 */
METHOD XbpGetColumn:init(oParent, oOwner, aPos, aSize, aPP, lVisible)
   ::xbpColumn:init(oParent, oOwner, aPos, aSize, aPP, lVisible)
   ::active :=.F.
   ::setHome:=.T.

   If oParent<>NIL .And. oParent:isDerivedFrom("XbpBrowse")
      ::browser:=oParent
   EndIf
Return Self


/*
 * Request system resources
 */
METHOD XbpGetColumn:create(oParent, oOwner, aPos, aSize, aPP, lVisible)
   LOCAL oError

   ::xbpColumn:create(oParent, oOwner, aPos, aSize, aPP, lVisible)

   /*
    * Set focus back to the browser If the column object gets input focus.
    * This assures that keyboard events are always processed by the browser.
    */
   ::dataArea:lbDblClick:={|aPos| ::_lbDblClick(aPos) }
   ::setInputFocus  :={|x,y,obj| SetAppFocus(obj:browser) }

   If ::browser == NIL
      /*
       * Find the XbpBrowse object in the Parent-Child hierarchy
       *
       * NOTE: The DO WHILE loop demonstrates how to search the parent-child
       *       hierarchy upwards. The top of the Parent-Child tree is reached
       *       when the Desktop window is the parent (oParent == AppDesktop())
       */
      oParent:=::setParent()

      DO WHILE .NOT. oParent:isDerivedFrom("XbpBrowse")
         oParent:=oParent:setParent()
         If oParent == AppDesktop()
            oError            :=Error():new()
            oError:description:="Illegal parent defined for " + ::className()
            oError:operation  :=ProcName()
            oError:args       :={ ::setParent() }
            Eval(ErrorBlock(), oError)
         EndIf
      ENDDO

      /*
       * An XbpBrowse object is found in the Parent-Child hierarchy
       */
      ::browser:=oParent
   EndIf
Return Self



/*
 * Clean up memory and set object references to NIL
 */
METHOD XbpGetColumn:destroy()
   If ::xbpGet<>NIL .And. ::xbpGet:status() > 0
      ::xbpGet:controller:=NIL
      ::xbpGet:destroy()
      ::xbpGet:=NIL
   EndIf
   ::active :=.F.
   ::browser:=NIL
   ::XbpColumn:destroy()
Return Self


/*
 * Create the XbpGet object and share the :dataLink code block
 */
METHOD XbpGetColumn:createXbp
   ::xbpGet           :=XbpGet():new(::dataArea, ::dataArea, NIL, NIL, NIL, .F.)
   ::xbpGet:controller:=Self
   ::xbpGet:dataLink  :=::dataLink

   If ::bufferLength<>NIL
      ::xbpGet:bufferLength:=::bufferLength
   EndIf

   If ::picture<>NIL
      ::xbpGet:picture:=::picture
   EndIf

   ::xbpGet:create()
Return


/*
 * Determine the X/Y coordinates for the XbpGet display.
 */
METHOD XbpGetColumn:setPosAndSize(nRowPos)
   LOCAL aRect, aPos, aSize

   aRect:=::dataArea:cellRect(nRowPos)
   aPos :={ aRect[1], aRect[2] }
   aSize:={ aRect[3]-aRect[1], aRect[4]-aRect[2] }
   ::xbpGet:setPos (aPos)
   ::xbpGet:setSize(aSize)
Return Self


/*
 * When the column is resized while Edit mode is active,
 * the edit control must be adjusted to the column width.
 */
METHOD XbpGetColumn:resize(aOldSize, aNewSize)

   If .NOT. ::active
      Return ::xbpColumn:resize(aOldSize, aNewSize)
   EndIf

   ::xbpGet:hide()
   ::xbpColumn:resize(aOldSize, aNewSize)
   ::setPosAndSize(::browser:rowPos)
   ::xbpGet:show()

   SetAppFocus(::xbpGet)
Return Self


/*
 * This method is called from the XbpBrowse object when the browse cursor
 * is moved. It must be overloaded for editable column objects and is
 * used to position the edit control at the browse cursor.
 */
METHOD XbpGetColumn:hiliteRow(nRowPos, lHilite, lFrame, lRepaint)

Begin Sequence

   If ::active
      If lHilite
         /*
          * We are in Edit mode and the browser wants to display its
          * cursor in highlight color. Resume with editing.
          */
         ::read(nRowPos)
      Else
         /*
          * We are in Edit mode and the browser wants to display its cursor
          * in normal color. Eval the :dataLink code block upon changes
          * and hide the XbpGet.
          */
         If ::xbpGet<>NIL
            If ::xbpGet:changed
               ::xbpGet:getData()
               ::xbpColumn:refreshRows(nRowPos, nRowPos)
            EndIf
            ::xbpGet:hide()
         EndIf
         ::xbpColumn:hiliteRow(nRowPos, lHilite, lFrame, lRepaint)
      EndIf
   Else
      /*
       * We are in non-edit mode. Pass everything on to the super class.
       */
      ::xbpColumn:hiliteRow(nRowPos, lHilite, lFrame, lRepaint)
   EndIf

End

Return Self


/*
 * Create and/or display the XbpGet object and edit the current row
 */
METHOD XbpGetColumn:read(nRowPos, nFirstKey)

Begin Sequence
   DEFAULT nRowPos TO ::browser:rowPos

   If ::activeColumn<>NIL  .And. ;
      ::activeColumn<>Self
      /*
       * The browser is clicked with the mouse while another column
       * is currently being edited. Terminate edit in that column.
       */
      ::activeColumn:killRead()
   EndIf

   ::activeColumn:=Self

   If ::xbpGet == NIL
      /*
       * The column has not been edited yet. Create the XbpGet object,
       * upon first editing.
       */
      ::createXbp()
   EndIf

   ::active :=.T.
   ::setHome:=.T.
   ::setPosAndSize(nRowPos)

   ::xbpGet:setData()
   ::xbpGet:changed(.F.)
   ::xbpGet:show()

   SetAppFocus(::xbpGet)

   If ::xbpGet:isDerivedFrom("XbpSLE")
      ::xbpGet:setMarked({1,1})
   EndIf

   If nFirstKey<>NIL
      ::xbpGet:clear()
      PostAppEvent(xbeP_Keyboard, nFirstKey, NIL, ::xbpGet)
   EndIf

End

Return Self


/*
 * Left button is doubleclicked in the :dataArea of the column
 */
METHOD XbpGetColumn:_lbDblClick(aPos)
   LOCAL nRowPos:=::dataArea:cellFromPos(aPos)

   ::active:=.T.

   If nRowPos == ::browser:rowPos
      ::read(nRowPos)
   EndIf
Return self


/*
 * Keyboard handling method
 *
 * When edit mode is active, the XbpGet object has input focus and
 * calls this method since the XbpColumn object is the controller
 * of the XbpGet object
 * (see method :createXbp() -> ::xbpGet:controller:=Self)
 *
 * - The edit mode is activated by
 *   1) the ENTER key (cell content remains)
 *   2) an editing key (the cell is cleared)
 *
 * - Esc terminates editing
 *
 */
METHOD XbpGetColumn:keyBoard(nKey)
   LOCAL bBlock

   DO CASE
   CASE nKey == xbeK_ALT_F4
      PostAppEvent(xbeP_Close,,, SetAppWindow())
      Return .T.

   CASE nKey == xbeK_ENTER
      /*
       * Edit next cell or activate edit mode.
       */
      If ::active

         If ::browser:colPos == ::browser:colCount
               bBlock:= {|o| ::browser:down():forceStable(), ;
                  ::browser:firstCol():right():forceStable(), ;
                  PostAppEvent( xbeP_Keyboard, xbeK_ENTER,, o ) }

               /*  old navi block
               bBlock:={|o| o:firstCol(), o:down(), o:right(), ;
                           PostAppEvent(xbeP_Keyboard, xbeK_ENTER,, o) }
               */
         Else
            bBlock:={|o| o:right(), ;
                           PostAppEvent(xbeP_Keyboard, xbeK_ENTER,, o) }
         EndIf

      Else
         ::read(::browser:rowPos)
      EndIf

   CASE .NOT. ::active
      /*
       * Keys processed by XbpGet activate edit mode and clear the
       * edit buffer. Other keys are ignored.
       */
      If nKey > 31 .And. nKey < 256
         ::read(::browser:rowPos, nKey)
      EndIf

   CASE nKey == xbeK_ESC
      /*
       * Undo changes upon ESC
       */
      ::xbpGet:undo()
      ::killRead()
      bBlock:={|o| o:refreshCurrent() }

   /*
    * Keys causing browse cursor movement during editing
    *
    * NOTE: Tab keys cause xbeK_ENTER to be posted to the browser so that
    *       the Edit mode is switched on in the next or previous column.
    */
   CASE nkey == xbeK_TAB

      If ::browser:colPos == ::browser:colCount
         bBlock:={|o| o:firstCol(), o:down(), o:right(),  ;
                        PostAppEvent(xbeP_Keyboard, xbeK_ENTER,, o) }
      Else
         bBlock:={|o| o:right(), ;
                        PostAppEvent(xbeP_Keyboard, xbeK_ENTER,, o) }
      EndIf

   CASE nkey == xbeK_SH_TAB
      If ::browser:colPos == 1
         bBlock:={|o| o:up(), o:lastCol(), ;
                        PostAppEvent(xbeP_Keyboard, xbeK_ENTER,, o) }
      Else
         bBlock:={|o| o:left(), ;
                        PostAppEvent(xbeP_Keyboard, xbeK_ENTER,, o) }
      EndIf

   CASE nkey == xbeK_DOWN
      bBlock:={|o| o:down() }

   CASE nKey == xbeK_UP
      bBlock:={|o| o:up() }

   CASE nkey == xbeK_PGDN
      bBlock:={|o| o:pageDown() }

   CASE nKey == xbeK_PGUP
      bBlock:={|o| o:pageUp() }

   CASE nkey == xbeK_CTRL_PGDN
      bBlock:={|o| o:goBottom() }

   CASE nKey == xbeK_CTRL_PGUP
      bBlock:={|o| o:goTop() }

   ENDCASE

   If bBlock<>NIL
      /*
       * Invoke browser action
       */
      Eval(bBlock, ::browser)
      ::browser:forceStable()
   EndIf

   /*
    * XbpGet needs a logical Return value
    */
Return (bBlock<>NIL)


/*
 * Terminate the Edit mode
 */
METHOD XbpGetColumn:killRead

   If ::active .And. ::postValidate()
      ::xbpGet:hide()
      ::active      :=.F.
      ::activeColumn:=NIL
      SetAppFocus(::browser)
   EndIf

Return Self
